
/*
 * cinecore.c, cinedbg.c, cineops.c -- Cinematronics game emulator
 * copyright 1997/1998 Jeff Mitchell and the Retrocade Alliance
 * copyright 1997 Zonn Moore
 *
 * This Cinematronics emulator may be freely used in any non-commercial
 * venture, and is to be considered as under the Gnu Public Licence
 * (GPL, copyleft). To avoid a huge deluge of email, I also allow the
 * MAME team usage under the so-called MAME licence for non-commercial
 * use.
 *
 * There are some restrictions, however, mostly to help further development
 * of the emulator and continue the enjoyment these fine games have
 * provided.
 *
 * 1) Jeff Mitchell (skeezix@skeleton.org) is the authoritative maintainer for
 *    the C implementation of the "CCPU" written by Zonn Moore. Modifications
 *    or changes to this code may not be distributed in source form, in whole
 *    or in part, without the written permission of the maintainer. (ie: At
 *    some point after changes have been made, submit them to the maintainer
 *    for inclusion into the authoritative source tree, for all to use.)
 * 2) Credit must be given where appropriate: Jeff Mitchell, author of the
 *    C version. Zonn Moore for the ASM version and all original research,
 *    original documentation, etc. (The C version is a rewrite of Zonn's
 *    brilliant work.) Neil Bradley, for the 32bit conversion of the ASM
 *    core, and technical assistance. Without him, the C version wouldn't
 *    have been written. Where possible, please include references to the
 *    official website for the software (see below), so that others may
 *    easily find the sources.
 * 3) Users of this software are encouraged to periodically check the
 *    official website for new revisions, bugfixes, etc. This is an
 *    ongoing project, with many optimizations yet to be done.
 *
 * Games known to work 100% or nearly so, at some point or another of
 * the emulators development (but not necessarily right now :):
 * RipOff, Solar Quest, Spacewar, Speed Freak, Star Castle, Star Hawk,
 * Tail Gunner, War of the Worlds, Warrior
 *
 * For reference, all of the cinematronics games are:
 * Armor Attack, Barrier, Boxing Bugs, Demon, Ripoff, Solar Quest,
 * Spacewar, Speed Freak, Star Castle, Star Hawk, Sundance, Tail Gunner
 * War of the worlds, Warrior
 * (There is another, but it has not been made available yet)
 *
 * USAGE:
 * 1) The emulator works like any other "processor" emulator
 * 2) It does expect a few variables to be set, however, indicating
 *    which switches were connected on the original boards. (Most
 *    important is the variable to chose whethor or not to use the MI
 *    or EI flag...). See below.
 * 3) The emulator expects the memory to ALREADY be interleaved into
 *    "normalacy". The cinem ROMs are slow to read otherwise...
 * 4) Endianness doesn't matter; its code nice and slowly :)
 * 5) Compile time options may be set to obtain debugging tool, runtime
 *    traces, etc etc.
 * 6) The emulator does the vector handling; you need provide a routine
 *    to receive the vector draw requests, so that it may queue them up
 * 7) Zonn will be documenting the method by which the game palettes
 *    may be derived, where applicable.
 *
 * INITIALIZATION:
 * Take, for example, the game RipOff (the reaosn I got into this mess :).
 * The following variables need to be set, and the functions called, in
 * order to have the "cpu" run properly (courtesy Neil Bradley, Zonn Moore):
 * bFlipX = 0;  bFlipY = 1;     bSwapXY = 0;     (for vector generation)
 * ioInputs = 0xffff; ioSwitches = 0xfff0;  (cleared values)
 * bOverlay = 0xff;                                 (No overlay)
 * cineSetJMI(1);                                      (JMI Allowed)
 * cineSetMSize(0);                            (8K)
 * cineSetMonitor(0);                       (Bi-level display)
 *
 * If masking values are needed by anyone for the various game
 * controllers and whatnot, let me know, and I'll put up a document.
 *
 * SUPPORT:
 * FTP and WEB:
 *   Official support archive is at ftp.retrocade.com and www.retrocade.com.
 *   Various mirrors will be kept (ie: skeleton.org, the Emulation Repository,
 *   erc.), and are yet to be announced.
 * EMAIL:
 *   The C version: Jeff Mitchell is skeezix@skeleton.org
 *   The ASM version: Neil Bradley (neil@synthcom.com) and
 *                    Zonn Moore (zonn@zonn.com)
 *
 * An emulator for the Cinematronics vector game hardware, originally
 * for use in the Retrocade emulator project. All work is based on
 * documentation and sources and testbeds created by Zonn. He's a freak.
 *
 * Last modified: 02/17/98
 *
 * 12/04/97:      Created a "SETFC" macro for storing the flag_C var
 *                (shift and mask it at inspection time)
 * 12/??/97:      Fixed subtraction bugs (using register_B for A0 instead
 *                of register_A). Fixed input mask value in opINP.
 * 12/24/97:      Added #define for CCPUSSTEP for single stepping
 * 12/25/97:      Added ioSwitches &= ( ! SW_ABORT ) into the top
 *                of the cineExec call. This fixed Star Castle, but
 *                broke others :/
 *                  Made opLSLe_A_.. handle multiple consecutive occurances
 *                so things would be a little faster, and so DOS versus
 *                Unix traces would be easier.
 *                  Changed above ioSwitches fix to ~SW_ABORT instead. This
 *                rebroke StarCastle and Sundance, but fixed Barrier. *sigh*
 */

/* Optional #defines for debugging:
 * CCPUBREAK -- causes stopage on unknown upcodes
 * CCPUDEBUG -- causes live trace to be dumped
 * For these to operate, the debugging variables must be set as well.
 * This allows runtime selection of traces, etc.
 * CCPUSSTEP -- force single stepping
 *
 * Debug Variables:
 * ccpubreak -- set to non-zero to enable "break"ing
 * ccpudebug -- set to non-zero to enable disassembly tracing
 */

/* 1) Need to macro-ize everything, so that I can have this whole
 *    source file written by a perl script generator easier.
 * 4) Any of the jumps weird? JNC?
 * 5) JEI's all fucked? Are the tredirector's set right in the first place?
 *    What about all those damned JPP8, 16 and 32s? They work right?
 * 6) Store ccpu_jmi_dip and other new state vars in struct?
 * 7) Various OUT functions correct?
 */

#include "StarterRsc.h"
#include "types.h"
#include "sim.h"
#include "ccpu.h"

// #define CCPUBREAK 1
// #define CCPUDEBUG 1

/* Declare needed macros
 */
#define UNFINISHED(x) fprintf ( stderr, "UNFINISHED: %s\n", x );

/* state setting macros to exit opcode handler */
#define jumpCineRet_A state = state_A; \
		      goto cineExecBottom;

#define jumpCineRet_B state = state_B; \
		      goto cineExecBottom;

#define jumpCineRet_AA state = state_AA; \
		       goto cineExecBottom;

#define jumpCineRet_BB state = state_BB; \
		       goto cineExecBottom;

#define jumpCineRet_WEIRD fprintf ( stderr, "Didn't bother yet.\n" ); \
			  exit ( 0 );

/* Handy new operators ...
 */

/* for Zonn translation :) */
#define SAR(var,arg)    ( ( (signed short int) var ) >> arg )

/* for setting/checking the A0 flag */
#define SETA0(var)    ( acc_a0 = var )
#define GETA0()       ( acc_a0 )

/* for setting/checking the Carry flag */
#define SETFC(val)    ( flag_C = val )
#define GETFC()       ( ( flag_C >> 8 ) & 0xFF )

/* Define new types for the c-cpu emulator
 */

#define SW_ABORT           0x100          /* for ioSwitches */
#define SW_COIN            0x080          /* for ioSwitches */

typedef UInt16 CINEWORD;      /* 12bits on the C-CPU */
typedef UInt8  CINEBYTE;      /* 8 (or less) bits on the C-CPU */
typedef Int16   CINESWORD;     /* 12bits on the C-CPU */
typedef Int8        CINESBYTE;     /* 8 (or less) bits on the C-CPU */

typedef enum {
  state_A = 0, state_AA, state_B, state_BB
} CINESTATE;                              /* current state */

volatile static UINT32 dwElapsedTicks = 0;
static UINT8 bBailOut = FALSE;

/* C-CPU context information begins --
 */
CINEWORD register_PC = 0; /* C-CPU registers; program counter */
CINEWORD register_A = 0;  /* ; A-Register (accumulator) */
CINEWORD register_B = 0;  /* ; B-Register (accumulator) */
CINEBYTE register_I = 0;  /* ; I-Register (last access RAM location) */
CINEWORD register_J = 0;  /* ; J-Register (target address for JMP opcodes) */
CINEBYTE register_P = 0;  /* ; Page-Register (4 bits) */
CINEWORD FromX = 0;       /* ; X-Register (start of a vector) */
CINEWORD FromY = 0;       /* ; Y-Register (start of a vector) */
CINEWORD register_T = 0;  /* ; T-Register (vector draw length timer) */
CINEWORD flag_C = 0;      /* C-CPU flags; carry. Is word sized, instead
			   * of CINEBYTE, so we can do direct assignment
			   * and then change to BYTE during inspection.
			   */

CINEWORD cmp_old = 0;     /* last accumulator value */
CINEWORD cmp_new = 0;     /* new accumulator value */
CINEBYTE acc_a0 = 0;      /* bit0 of A-reg at last accumulator access */

CINESTATE state = state_A; /* C-CPU state machine current state */
CINEWORD ram [ 256 ];     /* C-CPU ram (for all pages) */
UINT8 *rom = NULL; /* C-CPU rom code */

static UINT8 ccpu_jmi_dip = 0;  /* as set by cineSetJMI */
static UINT8 ccpu_msize = 0;    /* as set by cineSetMSize */
static UINT8 ccpu_monitor = 0;  /* as set by cineSetMonitor */

/* cinem graphics junk */
CINEBYTE vgShiftLength = 0; /* number of shifts loaded into length reg */

/* -- Context information ends.
 */

UINT8 bFlipX;
UINT8 bFlipY;
UINT8 bSwapXY;
UINT8 bOverlay;

UINT16 ioSwitches = 0xFF;  /* high values for non-triggered switches */
UINT16 ioInputs = 0;
UINT8 ioOutputs = 0;

UINT16 vgColour = 0;

INT16 JoyX;
INT16 JoyY;
UINT8 bNewFrame;

INT32 sdwGameXSize;
INT32 sdwGameYSize;
INT32 sdwXOffset;
INT32 sdwYOffset;

/* debug junk
 */
#include "cinedbg.c"

#ifdef CCPUBREAK
int ccpubreak = 1; /* default is on */
#endif

#ifdef CCPUDEBUG
int ccpudebug = 0; /* default is off */
#endif

/* functions
 */

/* cineExec() is what performs all the "processors" work; it will
 * continue to execute until something horrible happens, a watchpoint
 * is hit, cycle count exceeded, or other happy things.
 */
UINT32 cineExec ( UINT32 a ) {
#ifdef CCPUDEBUG
  char disassembly [ 256 ]; /* debugger buffer */
#endif
#ifdef CCPUBREAK
  CINEWORD original_PC;     /* before opcode begins executing */
  int original_state = 0;   /* before opcode executes */
  int repeat_count = 0;     /* each time PC repeats itself.. */
#endif

  CINEBYTE temp_byte = 0;   /* cause it'll be handy */
  CINEBYTE temp_byte_2 = 0; /* cause it'll be handy */
  CINEWORD temp_word = 0;   /* cause it'll be handy */
  CINEWORD temp_word_2 = 0; /* cause it'll be handy */

  CINESWORD temp_sword = 0; /* in case its handy... */
  CINESBYTE temp_sbyte = 0; /* in case its handy... */

  /* before opcode handlers ...
   */
  ioSwitches &= ( ~SW_ABORT ); /* ( ! SW_ABORT ) ? */

  /* label; after each opcode is finished running, execution returns
   * to this point, where the state is examined and the proper
   * jump table used.
   */
cineExecTop:

  /* back up the PC, for debugger sake
   */
#ifdef CCPUBREAK
  original_PC = register_PC;
#endif
#ifdef CCPUDEBUG
  original_state = state;
#endif

  bBailOut = FALSE;

#if 0
  /****/ /* for debugging only! */
  if ( register_PC == 0xabf ) {
    printf ( "got it: register_A = %x register_B = %x register_PC = %x\n",
	     register_A, register_B, register_PC );
  }
  /****/
#endif

  /* examine the current state, and jump down to the correct
   * opcode jump table.
   */
  switch ( state ) {
  case state_A:  goto opCodeTblA;
  case state_AA: goto opCodeTblAA;
  case state_B:  goto opCodeTblB;
  case state_BB: goto opCodeTblBB;
  } /* switch on current state */

  /* include the opcode jump tables; goto the correct piece of code
   * for the current opcode. That piece of code will set the state
   * for the next run, as well. The opcode jumptables were generated
   * with a perl script and are kept separate for make and ugly sake.
   */

#include "cineops.c"

  /* the actual opcode code; each piece should be careful to
   * (1) set the correct state
   * (2) increment the program counter as necessary
   * (3) piss with the flags as needed
   * otherwise the next opcode will be completely buggered.
   */

opINP_A_AA:
opINP_AA_AA:
opINP_BB_AA:
  /* bottom 4 bits of opcode are the position of the bit we want;
   * obtain input value, shift over that no, and truncate to last bit.
   * NOTE: Masking 0x07 does interesting things on Sundance and
   * others, but masking 0x0F makes RipOff and others actually work :)
   */
  cmp_new = ( ioInputs >> ( rom [ register_PC ] & 0x0F ) ) & 0x01;

  SETA0 ( register_A );               /* save old accA bit0 */
  SETFC ( register_A );

  cmp_old = register_A;               /* save old accB */
  register_A = cmp_new;               /* load new accB; zero other bits */

  register_PC ++;
  jumpCineRet_AA;

opINP_B_AA:
  /* bottom 3 bits of opcode are the position of the bit we want;
   * obtain Switches value, shift over that no, and truncate to last bit.
   */
  cmp_new = ( ioSwitches >> ( rom [ register_PC ] & 0x07 ) ) & 0x01;

  SETA0 ( register_A );               /* save old accA bit0 */
  SETFC ( register_A );

  cmp_old = register_B;               /* save old accB */
  register_B = cmp_new;               /* load new accB; zero other bits */

  register_PC ++;
  jumpCineRet_AA;

opOUTbi_A_A:
opOUTbi_AA_A:
opOUTbi_BB_A:

  temp_byte = rom [ register_PC ] & 0x07;
  register_PC++;

  if ( temp_byte - 0x06 ) {
    goto opOUTsnd_A;
  }

  vgColour = ( ( register_A & 0x01 ) << 3 ) | 0x07;

  jumpCineRet_A;

opOUT16_A_A:
opOUT16_AA_A:
opOUT16_BB_A:

  temp_byte = rom [ register_PC ] & 0x07;
  register_PC++;

  if ( temp_byte - 0x06 ) {
    goto opOUTsnd_A;
  }

  if ( ( register_A & 0xFF ) != 1 ) {
    vgColour = FromX & 0x0F;

    if ( ! vgColour ) {
      vgColour = 1;
    }

  }

  jumpCineRet_A;

opOUTsnd_A:
  temp_byte = 0x01 << ( rom [ register_PC ] & 0x07 );

  if ( ! ( register_A & 0x01 ) ) {
    goto opOUT_Aset;
  }

  temp_byte = ( ! temp_byte );
  ioOutputs &= temp_byte;

  if ( ( rom [ register_PC ] & 0x07 ) == 0x05 ) {
    goto opOUT_Aq;
  }

  jumpCineRet_A;

opOUT_Aq:
  /* reset coin counter */
  jumpCineRet_A;

opOUT_Aset:
  ioOutputs |= temp_byte;
  jumpCineRet_A;

opOUT64_A_A:
opOUT64_AA_A:
opOUT64_BB_A:
  jumpCineRet_A;

opOUTWW_A_A:
opOUTWW_AA_A:
opOUTWW_BB_A:
  temp_byte = rom [ register_PC ] & 0x07;
  register_PC++;

  if ( temp_byte - 0x06 ) {
    goto opOUTsnd_A;
  }

  if ( ( register_A & 0xFF ) == 1 ) {
    temp_word = ( ! FromX ) & 0x0FFF;

    if ( ! temp_word ) {   /* black */

      vgColour = 0;

    } else {              /* non-black */

      if ( temp_word & 0x0888 ) {
	/* bright */

	temp_word_2 = ( ( temp_word << 4 ) & 0x8000 ) >> 15;
	temp_byte = ( temp_byte << 1 ) + temp_word_2;

	temp_word_2 = ( ( temp_word << 3 ) & 0x8000 ) >> 15;
	temp_byte = ( temp_byte << 1 ) + temp_word_2;

	temp_word_2 = ( ( temp_word << 3 ) & 0x8000 ) >> 15;
	temp_byte = ( temp_byte << 1 ) + temp_word_2;

	vgColour = ( temp_byte & 0x07 ) + 7;

      } else if ( temp_word & 0x0444 ) {
	/* dim bits */

	temp_word_2 = ( ( temp_word << 5 ) & 0x8000 ) >> 15;
	temp_byte = ( temp_byte << 1 ) + temp_word_2;

	temp_word_2 = ( ( temp_word << 3 ) & 0x8000 ) >> 15;
	temp_byte = ( temp_byte << 1 ) + temp_word_2;

	temp_word_2 = ( ( temp_word << 3 ) & 0x8000 ) >> 15;
	temp_byte = ( temp_byte << 1 ) + temp_word_2;

	vgColour = ( temp_byte & 0x07 );

      } else {
	/* dim white */
	vgColour = 0x0F;

      }

    }

  } /* colour change? == 1 */

  jumpCineRet_A;

opOUTbi_B_BB:
  temp_byte = rom [ register_PC ] & 0x07;
  register_PC++;

  if ( temp_byte - 0x06 ) {
    goto opOUTsnd_B;
  }

  vgColour = ( ( register_B & 0x01 ) << 3 ) | 0x07;

  jumpCineRet_BB;

opOUT16_B_BB:
  temp_byte = rom [ register_PC ] & 0x07;
  register_PC++;

  if ( temp_byte - 0x06 ) {
    goto opOUTsnd_B;
  }

  if ( ( register_B & 0xFF ) != 1 ) {
    vgColour = FromX & 0x0F;

    if ( ! vgColour ) {
      vgColour = 1;
    }

  }

  jumpCineRet_BB;

opOUTsnd_B:
  jumpCineRet_BB;

opOUT64_B_BB:
  jumpCineRet_BB;

opOUTWW_B_BB:
  jumpCineRet_BB;

  /* LDA imm (0x) */
opLDAimm_A_AA:
opLDAimm_AA_AA:
opLDAimm_BB_AA:
  temp_word = rom [ register_PC ] & 0x0F;   /* pick up immediate value */
  temp_word <<= 8;                          /* LDAimm is the HIGH nibble! */

  cmp_new = temp_word;                      /* set new comparison flag */

  SETA0 ( register_A );                     /* save old accA bit0 */
  SETFC ( register_A );                     /* ??? clear carry? */

  cmp_old = register_A;                     /* step back cmp flag */
  register_A = temp_word;                   /* set the register */

  register_PC++;                            /* increment PC */
  jumpCineRet_AA;                           /* swap state and end opcode */

opLDAimm_B_AA:
  temp_word = rom [ register_PC ] & 0x0F;   /* pick up immediate value */
  temp_word <<= 8;                          /* LDAimm is the HIGH nibble! */

  cmp_new = temp_word;                      /* set new comparison flag */

  SETA0 ( register_A );                     /* save old accA bit0 */
  SETFC ( register_A );

  cmp_old = register_B;                     /* step back cmp flag */
  register_B = temp_word;                   /* set the register */

  register_PC++;                            /* increment PC */
  jumpCineRet_AA;

opLDAdir_A_AA:
opLDAdir_AA_AA:
opLDAdir_BB_AA:

  temp_byte = rom [ register_PC ] & 0x0F;        /* snag imm value */
  register_I = ( register_P << 4 ) + temp_byte;  /* set I register */

  cmp_new = ram[ register_I ];                  /* new acc value */

  SETA0 ( register_A );                          /* back up bit0 */
  SETFC ( register_A );

  cmp_old = register_A;                          /* store old acc */
  register_A = cmp_new;                          /* store new acc */

  register_PC ++;                                /* bump PC */
  jumpCineRet_AA;

opLDAdir_B_AA:

  temp_byte = rom [ register_PC ] & 0x0F;        /* snag imm value */
  register_I = ( register_P << 4 ) + temp_byte;  /* set I register */

  cmp_new = ram[ register_I ];                  /* new acc value */

  SETA0 ( register_A );                          /* back up bit0 */
  SETFC ( register_A );

  cmp_old = register_B;                          /* store old acc */
  register_B = cmp_new;                          /* store new acc */

  register_PC ++;                                /* bump PC */
  jumpCineRet_AA;

opLDAirg_A_AA:
opLDAirg_AA_AA:
opLDAirg_BB_AA:

  cmp_new = ram[ register_I ];

  SETA0 ( register_A );
  SETFC ( register_A );

  cmp_old = register_A;
  register_A = cmp_new;

  register_PC++;
  jumpCineRet_AA;

opLDAirg_B_AA:
  cmp_new = ram[ register_I ];

  SETA0 ( register_A );
  SETFC ( register_A );

  cmp_old = register_B;
  register_B = cmp_new;

  register_PC++;
  jumpCineRet_AA;

  /* ADD imm */
opADDimm_A_AA:
opADDimm_AA_AA:
opADDimm_BB_AA:
  temp_word = rom [ register_PC ] & 0x0F;     /* get imm value */

  cmp_new = temp_word;                        /* new acc value */
  SETA0 ( register_A );                       /* save old accA bit0 */
  cmp_old = register_A;                       /* store old acc for later */

  register_A += temp_word;                    /* add values */
  SETFC ( register_A );                       /* store carry and extra */
  register_A &= 0xFFF;                        /* toss out >12bit carry */

  register_PC ++;
  jumpCineRet_AA;

opADDimm_B_AA:
  temp_word = rom [ register_PC ] & 0x0F;     /* get imm value */

  cmp_new = temp_word;                        /* new acc value */
  SETA0 ( register_A );                       /* save old accA bit0 */
  cmp_old = register_B;                       /* store old acc for later */

  register_B += temp_word;                    /* add values */
  SETFC ( register_B );                       /* store carry and extra */
  register_B &= 0xFFF;                        /* toss out >12bit carry */

  register_PC ++;
  jumpCineRet_AA;

  /* ADD imm extended */
opADDimmX_A_AA:
opADDimmX_AA_AA:
opADDimmX_BB_AA:
  cmp_new = rom [ register_PC + 1 ];          /* get extended value */
  SETA0 ( register_A );                       /* save old accA bit0 */
  cmp_old = register_A;                       /* store old acc for later */

  register_A += cmp_new;                      /* add values */
  SETFC ( register_A );                       /* store carry and extra */
  register_A &= 0xFFF;                        /* toss out >12bit carry */

  register_PC += 2;                           /* bump PC */
  jumpCineRet_AA;

opADDimmX_B_AA:
  cmp_new = rom [ register_PC + 1 ];          /* get extended value */
  SETA0 ( register_A );                       /* save old accA bit0 */
  cmp_old = register_B;                       /* store old acc for later */

  register_B += cmp_new;                      /* add values */
  SETFC ( register_B );                       /* store carry and extra */
  register_B &= 0xFFF;                        /* toss out >12bit carry */

  register_PC += 2;                           /* bump PC */
  jumpCineRet_AA;

opADDdir_A_AA:
opADDdir_AA_AA:
opADDdir_BB_AA:

  temp_byte = rom [ register_PC ] & 0x0F;         /* fetch imm value */
  register_I = ( register_P << 4 ) + temp_byte;   /* set regI addr */

  cmp_new = ram[ register_I ];                   /* fetch imm real value */
  SETA0 ( register_A );                           /* store bit0 */
  cmp_old = register_A;                           /* store old acc value */

  register_A += cmp_new;                          /* do acc operation */
  SETFC ( register_A );                           /* store carry and extra */
  register_A &= 0x0FFF;

  register_PC ++;                                 /* bump PC */
  jumpCineRet_AA;

opADDdir_B_AA:
  temp_byte = rom [ register_PC ] & 0x0F;         /* fetch imm value */
  register_I = ( register_P << 4 ) + temp_byte;   /* set regI addr */

  cmp_new = ram[ register_I ];                   /* fetch imm real value */
  SETA0 ( register_A );                           /* store bit0 */
  cmp_old = register_B;                           /* store old acc value */

  register_B += cmp_new;                          /* do acc operation */
  SETFC ( register_B );                           /* store carry and extra */
  register_B &= 0x0FFF;

  register_PC ++;                                 /* bump PC */
  jumpCineRet_AA;

opAWDirg_A_AA:
opAWDirg_AA_AA:
opAWDirg_BB_AA:
opADDirg_A_AA:
opADDirg_AA_AA:
opADDirg_BB_AA:

  cmp_new = ram[ register_I ];
  SETA0 ( register_A );
  cmp_old = register_A;

  register_A += cmp_new;
  SETFC ( register_A );
  register_A &= 0x0FFF;

  register_PC++;
  jumpCineRet_AA;

opAWDirg_B_AA:
opADDirg_B_AA:
  cmp_new = ram[ register_I ];
  SETA0 ( register_A );
  cmp_old = register_B;

  register_B += cmp_new;
  SETFC ( register_B );
  register_B &= 0x0FFF;

  register_PC++;
  jumpCineRet_AA;

opSUBimm_A_AA:
opSUBimm_AA_AA:
opSUBimm_BB_AA:
  /* SUBtractions are negate-and-add instructions of the CCPU; what
   * a pain in the ass.
   */
  temp_word = rom [ register_PC ] & 0x0F;

  cmp_new = temp_word;
  SETA0 ( register_A );
  cmp_old = register_A;

  temp_word = ( temp_word ^ 0xFFF ) + 1;         /* ones compliment */
  register_A += temp_word;                       /* add */
  SETFC ( register_A );                          /* pick up top bits */
  register_A &= 0x0FFF;                          /* mask final regA value */

  register_PC ++;                                /* bump PC */
  jumpCineRet_AA;

opSUBimm_B_AA:
  /* SUBtractions are negate-and-add instructions of the CCPU; what
   * a pain in the ass.
   */
  temp_word = rom [ register_PC ] & 0x0F;

  cmp_new = temp_word;
  SETA0 ( register_A );
  cmp_old = register_B;

  temp_word = ( temp_word ^ 0xFFF ) + 1;         /* ones compliment */
  register_B += temp_word;                       /* add */
  SETFC ( register_B );                          /* pick up top bits */
  register_B &= 0x0FFF;                          /* mask final regA value */

  register_PC ++;                                /* bump PC */
  jumpCineRet_AA;

opSUBimmX_A_AA:
opSUBimmX_AA_AA:
opSUBimmX_BB_AA:

  temp_word = rom [ register_PC + 1 ];          /* snag imm value */

  cmp_new = temp_word;                          /* save cmp value */
  SETA0 ( register_A );                         /* store bit0 */
  cmp_old = register_A;                         /* back up regA */

  temp_word = ( temp_word ^ 0xFFF ) + 1;        /* ones compliment */
  register_A += temp_word;                      /* add */
  SETFC ( register_A );                         /* pick up top bits */
  register_A &= 0x0FFF;                         /* mask final regA value */

  register_PC += 2;                             /* bump PC */
  jumpCineRet_AA;

opSUBimmX_B_AA:

  temp_word = rom [ register_PC + 1 ];          /* snag imm value */

  cmp_new = temp_word;                          /* save cmp value */
  SETA0 ( register_A );                         /* store bit0 */
  cmp_old = register_B;                         /* back up regA */

  temp_word = ( temp_word ^ 0xFFF ) + 1;        /* ones compliment */
  register_B += temp_word;                      /* add */
  SETFC ( register_B );                         /* pick up top bits */
  register_B &= 0x0FFF;                         /* mask final regA value */

  register_PC += 2;                             /* bump PC */
  jumpCineRet_AA;

opSUBdir_A_AA:
opSUBdir_AA_AA:
opSUBdir_BB_AA:
  temp_word = rom [ register_PC ] & 0x0F;         /* fetch imm value */
  register_I = ( register_P << 4 ) + temp_word;   /* set regI addr */

  cmp_new = ram[ register_I ];
  SETA0 ( register_A );
  cmp_old = register_A;

  temp_word = ( cmp_new ^ 0xFFF ) + 1;           /* ones compliment */
  register_A += temp_word;                       /* add */
  SETFC ( register_A );                          /* pick up top bits */
  register_A &= 0x0FFF;                          /* mask final regA value */

  register_PC ++;
  jumpCineRet_AA;

opSUBdir_B_AA:

  temp_byte = rom [ register_PC ] & 0x0F;         /* fetch imm value */
  register_I = ( register_P << 4 ) + temp_byte;   /* set regI addr */

  cmp_new = ram[ register_I ];
  SETA0 ( register_A );
  cmp_old = register_B;

  temp_word = ( cmp_new ^ 0xFFF ) + 1;           /* ones compliment */
  register_B += temp_word;                       /* add */
  SETFC ( register_B );                          /* pick up top bits */
  register_B &= 0x0FFF;                          /* mask final regA value */

  register_PC ++;
  jumpCineRet_AA;

opSUBirg_A_AA:
opSUBirg_AA_AA:
opSUBirg_BB_AA:
  /* sub [i] */
  cmp_new = ram[ register_I ];
  SETA0 ( register_A );
  cmp_old = register_A;

  temp_word = ( cmp_new ^ 0xFFF ) + 1;           /* ones compliment */
  register_A += temp_word;                       /* add */
  SETFC ( register_A );                          /* pick up top bits */
  register_A &= 0x0FFF;                          /* mask final regA value */

  register_PC ++;
  jumpCineRet_AA;

opSUBirg_B_AA:
  /* sub [i] */
  cmp_new = ram[ register_I ];
  SETA0 ( register_A );
  cmp_old = register_B;

  temp_word = ( cmp_new ^ 0xFFF ) + 1;           /* ones compliment */
  register_B += temp_word;                       /* add */
  SETFC ( register_B );                          /* pick up top bits */
  register_B &= 0x0FFF;                          /* mask final regA value */

  register_PC ++;
  jumpCineRet_AA;

  /* CMP dir */
opCMPdir_A_AA:
opCMPdir_AA_AA:
opCMPdir_BB_AA:
  /* compare direct mode; don't modify regs, just set carry flag or not.
   */

  temp_byte = rom [ register_PC ] & 0x0F;       /* obtain relative addr */
  register_I = ( register_P << 4 ) + temp_byte; /* build real addr */

  temp_word = ram[ register_I ];
  cmp_new = temp_word;                          /* new acc value */
  SETA0 ( register_A );                         /* backup bit0 */
  cmp_old = register_A;                         /* backup old acc */

  temp_word = ( temp_word ^ 0xFFF ) + 1;        /* ones compliment */
  temp_word += register_A;
  SETFC ( temp_word );                          /* pick up top bits */

  register_PC ++;                               /* bump PC */
  jumpCineRet_AA;

opCMPdir_B_AA:
  temp_byte = rom [ register_PC ] & 0x0F;       /* obtain relative addr */
  register_I = ( register_P << 4 ) + temp_byte; /* build real addr */

  temp_word = ram[ register_I ];
  cmp_new = temp_word;                          /* new acc value */
  SETA0 ( register_A );                         /* backup bit0 */
  cmp_old = register_B;                         /* backup old acc */

  temp_word = ( temp_word ^ 0xFFF ) + 1;        /* ones compliment */
  temp_word += register_B;
  SETFC ( temp_word );                          /* pick up top bits */

  register_PC ++;                               /* bump PC */
  jumpCineRet_AA;

  /* AND [i] */
opANDirg_A_AA:
opANDirg_AA_AA:
opANDirg_BB_AA:
  cmp_new = ram[ register_I ];                /* new acc value */
  SETA0 ( register_A );
  SETFC ( register_A );
  cmp_old = register_A;

  register_A &= cmp_new;

  register_PC ++;                              /* bump PC */
  jumpCineRet_AA;

opANDirg_B_AA:
  cmp_new = ram[ register_I ];                /* new acc value */
  SETA0 ( register_A );
  SETFC ( register_A );
  cmp_old = register_B;

  register_B &= cmp_new;

  register_PC ++;                              /* bump PC */
  jumpCineRet_AA;

  /* LDJ imm */
opLDJimm_A_A:
opLDJimm_AA_A:
opLDJimm_BB_A:
  temp_byte = rom [ register_PC + 1 ];         /* upper part of address */
  temp_byte = ( temp_byte << 4 ) |             /* Silly CCPU; Swap */
	      ( temp_byte >> 4 );              /* nibbles */

  /* put the upper 8 bits above the existing 4 bits */
  register_J = ( rom [ register_PC ] & 0x0F ) | ( temp_byte << 4 );

  register_PC += 2;
  jumpCineRet_A;

opLDJimm_B_BB:
  temp_byte = rom [ register_PC + 1 ];         /* upper part of address */
  temp_byte = ( temp_byte << 4 ) |             /* Silly CCPU; Swap */
	      ( temp_byte >> 4 );              /* nibbles */

  /* put the upper 8 bits above the existing 4 bits */
  register_J = ( rom [ register_PC ] & 0x0F ) | ( temp_byte << 4 );

  register_PC += 2;
  jumpCineRet_BB;

  /* LDJ irg */
opLDJirg_A_A:
opLDJirg_AA_A:
opLDJirg_BB_A:
  /* load J reg from value at last dir addr */
  register_J = ram[ register_I ];
  register_PC++;                             /* bump PC */
  jumpCineRet_A;

opLDJirg_B_BB:
  register_J = ram[ register_I ];
  register_PC++;                             /* bump PC */
  jumpCineRet_BB;

  /* LDP imm */
opLDPimm_A_A:
opLDPimm_AA_A:
opLDPimm_BB_A:
  /* load page register from immediate */
  register_P = rom [ register_PC ] & 0x0F;  /* set page register */
  register_PC ++;  /* inc PC */
  jumpCineRet_A;

opLDPimm_B_BB:
  /* load page register from immediate */
  register_P = rom [ register_PC ] & 0x0F;  /* set page register */
  register_PC ++;  /* inc PC */
  jumpCineRet_BB;

  /* LDI dir */
opLDIdir_A_A:
opLDIdir_AA_A:
opLDIdir_BB_A:
  /* load regI directly .. */

  temp_byte = ( register_P << 4 ) +           /* get rampage ... */
	      ( rom [ register_PC ] & 0x0F ); /* and imm half of ramaddr.. */

  register_I = ram[ temp_byte ] & 0xFF;      /* set/mask new register_I */

  register_PC ++;
  jumpCineRet_A;

opLDIdir_B_BB:
  temp_byte = ( register_P << 4 ) +           /* get rampage ... */
	      ( rom [ register_PC ] & 0x0F ); /* and imm half of ramaddr.. */

  register_I = ram[ temp_byte ] & 0xFF;      /* set/mask new register_I */

  register_PC ++;
  jumpCineRet_BB;

  /* STA dir */
opSTAdir_A_A:
opSTAdir_AA_A:
opSTAdir_BB_A:
  temp_byte = rom [ register_PC ] & 0x0F;        /* snag imm value */
  register_I = ( register_P << 4 ) + temp_byte;  /* set I register */

  ram[ register_I ] = register_A;               /* store acc to ram*/

  register_PC ++;                                /* inc PC */
  jumpCineRet_A;

opSTAdir_B_BB:
  temp_byte = rom [ register_PC ] & 0x0F;        /* snag imm value */
  register_I = ( register_P << 4 ) + temp_byte;  /* set I register */

  ram[ register_I ] = register_B;               /* store acc to ram*/

  register_PC ++;                                /* inc PC */
  jumpCineRet_BB;

  /* STA irg */
opSTAirg_A_A:
opSTAirg_AA_A:
opSTAirg_BB_A:
  /* STA into address specified in regI. Nice and easy :)
   */

  ram[ register_I ] = register_A;               /* store acc */

  register_PC ++;                                /* bump PC */
  jumpCineRet_A;

opSTAirg_B_BB:

  ram[ register_I ] = register_B;               /* store acc */

  register_PC ++;                                /* bump PC */
  jumpCineRet_BB;

  /* XLT */
opXLT_A_AA:
opXLT_AA_AA:
opXLT_BB_AA:
  /* XLT is weird; it loads the current accumulator with the bytevalue
   * at ROM location pointed to by the accumulator; this allows the
   * programto read the programitself..
   *   NOTE! Next opcode is *IGNORED!* because of a twisted side-effect
   */

  cmp_new = rom [ ( register_PC & 0xF000 ) + register_A ];   /* store new acc value */
  SETA0 ( register_A );           /* store bit0 */
  SETFC ( register_A );
  cmp_old = register_A;           /* back up acc */

  register_A = cmp_new;           /* new acc value */

  register_PC += 2;               /* bump PC twice because XLT is fucked up */
  jumpCineRet_AA;

opXLT_B_AA:

  cmp_new = rom [ ( register_PC & 0xF000 ) + register_B ];   /* store new acc value */
  SETA0 ( register_A );           /* store bit0 */
  SETFC ( register_A );
  cmp_old = register_B;           /* back up acc */

  register_B = cmp_new;           /* new acc value */

  register_PC += 2;               /* bump PC twice because XLT is fucked up */
  jumpCineRet_AA;

  /* MUL [i] */
opMULirg_A_AA:
opMULirg_AA_AA:
opMULirg_BB_AA:
  /* MUL's usually happen in batches, so a slight speed bump can be
   * gained by checking for multiple instances and handling in here,
   * without going through the main loop for each.
   */
  temp_word = ram[ register_I ];               /* pick up ramvalue */
  cmp_new = temp_word;

  temp_word <<= 4;                              /* shift into ADD position */
  register_B <<= 4;                             /* get sign bit 15 */
  register_B |= ( register_A >> 8 );            /* bring in A high nibble */

  register_A = ( ( register_A & 0xFF ) << 8 ) | /* shift over 8 bits */
	       ( rom [ register_PC ] & 0xFF );  /* pick up opcode */

  temp_byte = rom [ register_PC ] & 0xFF;       /* (for ease and speed) */

  /* handle multiple consecutive MUL's
   */

  register_PC++;                               /* inc PC */
  if ( rom [ register_PC ] != temp_byte ) {    /* next opcode is a MUL? */
    goto opMUL1;                               /* no? skip multiples... */
  }

  register_PC++;                               /* repeat above */
  if ( rom [ register_PC ] != temp_byte ) {
    goto opMUL2;
  }

  register_PC++;                               /* repeat above */
  if ( rom [ register_PC ] != temp_byte ) {
    goto opMUL3;
  }

  register_PC++;                               /* repeat above */
  if ( rom [ register_PC ] != temp_byte ) {
    goto opMUL4;
  }

  register_PC++;                               /* repeat above */
  if ( rom [ register_PC ] != temp_byte ) {
    goto opMUL5;
  }

  register_PC++;                               /* repeat above */
  if ( rom [ register_PC ] != temp_byte ) {
    goto opMUL6;
  }

  register_PC++;                               /* repeat above */
  if ( rom [ register_PC ] != temp_byte ) {
    goto opMUL7;
  }

  register_PC++;                               /* repeat above */
  if ( rom [ register_PC ] != temp_byte ) {
    goto opMUL8;
  }

  register_PC++;                               /* repeat above */
  if ( rom [ register_PC ] != temp_byte ) {
    goto opMUL9;
  }

  register_PC++;                               /* repeat above */
  if ( rom [ register_PC ] != temp_byte ) {
    goto opMUL10;
  }

  register_PC++;                               /* repeat above */
  if ( rom [ register_PC ] != temp_byte ) {
    goto opMUL11;
  }

opMUL12:
  register_PC++;                               /* we don't bother to check
						* for multiple occurances more
						* than 12, so just inc the PC
						* and not worry about it.
						*/

  temp_word_2 = ( register_B & 0x01 ) << 15;   /* get low bit for rotation */
  register_B = SAR(register_B,1);              /* signed arith right 1 */
  register_A = ( register_A >> 1 ) |           /* rotate right ... */
	       temp_word_2;                    /*   ... via carry flag */
  if ( ! ( register_A & 0x80 ) ) {
    goto opMUL11;
  }
  register_B += temp_word;

opMUL11:
  temp_word_2 = ( register_B & 0x01 ) << 15;   /* get low bit for rotation */
  register_B = SAR(register_B,1);              /* signed arith right 1 */
  register_A = ( register_A >> 1 ) |           /* rotate right ... */
	       temp_word_2;                    /*   ... via carry flag */
  if ( ! ( register_A & 0x80 ) ) {
    goto opMUL10;
  }
  register_B += temp_word;

opMUL10:
  temp_word_2 = ( register_B & 0x01 ) << 15;   /* get low bit for rotation */
  register_B = SAR(register_B,1);              /* signed arith right 1 */
  register_A = ( register_A >> 1 ) |           /* rotate right ... */
	       temp_word_2;                    /*   ... via carry flag */
  if ( ! ( register_A & 0x80 ) ) {
    goto opMUL9;
  }
  register_B += temp_word;

opMUL9:
  temp_word_2 = ( register_B & 0x01 ) << 15;   /* get low bit for rotation */
  register_B = SAR(register_B,1);              /* signed arith right 1 */
  register_A = ( register_A >> 1 ) |           /* rotate right ... */
	       temp_word_2;                    /*   ... via carry flag */
  if ( ! ( register_A & 0x80 ) ) {
    goto opMUL8;
  }
  register_B += temp_word;

opMUL8:
  temp_word_2 = ( register_B & 0x01 ) << 15;   /* get low bit for rotation */
  register_B = SAR(register_B,1);              /* signed arith right 1 */
  register_A = ( register_A >> 1 ) |           /* rotate right ... */
	       temp_word_2;                    /*   ... via carry flag */
  if ( ! ( register_A & 0x80 ) ) {
    goto opMUL7;
  }
  register_B += temp_word;

opMUL7:
  temp_word_2 = ( register_B & 0x01 ) << 15;   /* get low bit for rotation */
  register_B = SAR(register_B,1);              /* signed arith right 1 */
  register_A = ( register_A >> 1 ) |           /* rotate right ... */
	       temp_word_2;                    /*   ... via carry flag */
  if ( ! ( register_A & 0x80 ) ) {
    goto opMUL6;
  }
  register_B += temp_word;

opMUL6:
  temp_word_2 = ( register_B & 0x01 ) << 15;   /* get low bit for rotation */
  register_B = SAR(register_B,1);              /* signed arith right 1 */
  register_A = ( register_A >> 1 ) |           /* rotate right ... */
	       temp_word_2;                    /*   ... via carry flag */
  if ( ! ( register_A & 0x80 ) ) {
    goto opMUL5;
  }
  register_B += temp_word;

opMUL5:
  temp_word_2 = ( register_B & 0x01 ) << 15;   /* get low bit for rotation */
  register_B = SAR(register_B,1);              /* signed arith right 1 */
  register_A = ( register_A >> 1 ) |           /* rotate right ... */
	       temp_word_2;                    /*   ... via carry flag */
  if ( ! ( register_A & 0x80 ) ) {
    goto opMUL4;
  }
  register_B += temp_word;

opMUL4:
  temp_word_2 = ( register_B & 0x01 ) << 15;   /* get low bit for rotation */
  register_B = SAR(register_B,1);              /* signed arith right 1 */
  register_A = ( register_A >> 1 ) |           /* rotate right ... */
	       temp_word_2;                    /*   ... via carry flag */
  if ( ! ( register_A & 0x80 ) ) {
    goto opMUL3;
  }
  register_B += temp_word;

opMUL3:
  temp_word_2 = ( register_B & 0x01 ) << 15;   /* get low bit for rotation */
  register_B = SAR(register_B,1);              /* signed arith right 1 */
  register_A = ( register_A >> 1 ) |           /* rotate right ... */
	       temp_word_2;                    /*   ... via carry flag */
  if ( ! ( register_A & 0x80 ) ) {
    goto opMUL2;
  }
  register_B += temp_word;

opMUL2:
  temp_word_2 = ( register_B & 0x01 ) << 15;   /* get low bit for rotation */
  register_B = SAR(register_B,1);              /* signed arith right 1 */
  register_A = ( register_A >> 1 ) |           /* rotate right ... */
	       temp_word_2;                    /*   ... via carry flag */
  if ( ! ( register_A & 0x80 ) ) {
    goto opMUL1;
  }
  register_B += temp_word;

opMUL1:

  if ( register_A & 0x100 ) {           /* 1bit shifted out? */
    goto opMULshf;
  }

  register_A = ( register_A >> 8 ) |    /* Bhigh | Alow */
	       ( ( register_B & 0xFF ) << 8 );

  temp_word = register_A & 0xFFF;

  SETA0 ( temp_word & 0xFF );                   /* store bit0 */
  cmp_old = temp_word;

  temp_word += cmp_new;
  SETFC ( temp_word );

  register_A >>= 1;
  register_A &= 0xFFF;

  register_B = SAR(register_B,5);
  register_B &= 0xFFF;

  jumpCineRet_AA;

opMULshf: /* part of opMULirg */

  register_A = ( register_A >> 8 ) |
	       ( ( register_B & 0xFF ) << 8 );

  SETA0 ( register_A & 0xFF );                  /* store bit0 */

  register_A >>= 1;
  register_A &= 0xFFF;

  register_B = SAR(register_B,4);
  cmp_old = register_B & 0x0F;

  register_B = SAR(register_B,1);

  register_B &= 0xFFF;
  register_B += cmp_new;

  SETFC ( register_B );

  register_B &= 0xFFF;

  jumpCineRet_AA;

opMULirg_B_AA:
  register_PC++;

  temp_word = ram[ register_I ];
  cmp_new = temp_word;
  cmp_old = register_B;
  SETA0 ( register_A & 0xFF );

  register_B <<= 4;

  register_B = SAR(register_B,5);

  if ( register_A & 0x01 ) {
    goto opMULirgB1;
  }

  temp_word += register_B;
  SETFC ( temp_word );

  jumpCineRet_AA;

opMULirgB1:
  register_B += temp_word;
  SETFC ( register_B );
  register_B &= 0x0FFF;
  jumpCineRet_AA;

  /* LSRe */
opLSRe_A_AA:
opLSRe_AA_AA:
opLSRe_BB_AA:
  /* EB; right shift pure; fill new bit with zero.
   */
  temp_word = 0x0BEB;

  register_PC++;

  if ( rom [ register_PC ] == 0xEB ) {
    goto opLSRe_A0; /* multiples */
  }

  cmp_new = temp_word;
  SETA0 ( register_A );
  cmp_old = register_A;

  temp_word += register_A;
  SETFC ( temp_word );

  register_A >>= 1;
  jumpCineRet_AA;

opLSRe_A0:
  register_A >>= 1;
  register_PC ++;
  if ( rom [ register_PC ] != 0xEB ) {
    goto opLSRe_A1;
  }

  register_A >>= 1;
  register_PC ++;
  if ( rom [ register_PC ] != 0xEB ) {
    goto opLSRe_A1;
  }

  register_A >>= 1;
  register_PC ++;
  if ( rom [ register_PC ] != 0xEB ) {
    goto opLSRe_A1;
  }

  register_A >>= 1;
  register_PC ++;
  if ( rom [ register_PC ] != 0xEB ) {
    goto opLSRe_A1;
  }

  register_A >>= 1;
  register_PC ++;
  if ( rom [ register_PC ] != 0xEB ) {
    goto opLSRe_A1;
  }

  register_A >>= 1;
  register_PC ++;
  if ( rom [ register_PC ] != 0xEB ) {
    goto opLSRe_A1;
  }

  register_A >>= 1;
  register_PC ++;

opLSRe_A1:
  cmp_new = temp_word;
  SETA0 ( register_A );
  cmp_old = register_A;

  temp_word += register_A;
  SETFC ( temp_word );

  register_A >>= 1;
  jumpCineRet_AA;

opLSRe_B_AA:
  UNFINISHED ( "opLSRe B 1\n" );
  jumpCineRet_AA;

opLSRf_A_AA:
opLSRf_AA_AA:
opLSRf_BB_AA:
  UNFINISHED ( "opLSRf 1\n" );
  jumpCineRet_AA;

opLSRf_B_AA:
  UNFINISHED ( "opLSRf 2\n" );
  jumpCineRet_AA;

opLSLe_A_AA:
opLSLe_AA_AA:
opLSLe_BB_AA:
  /* EC; left shift pure; fill new bit with zero */
  /* This version supports multiple consecutive LSLe's; the older
   * version only did one at a time. I'm changing it to make tracing
   * easier (as its comperable to Zonn's)
   */

  register_PC++;
  temp_word = 0x0CEC;

  if ( rom [ register_PC ] == 0xEC ) {
    goto opLSLe_A0; /* do multiples */
  }

  cmp_new = temp_word;
  SETA0 ( register_A );
  cmp_old = register_A;

  temp_word += register_A;
  SETFC ( temp_word );

  register_A <<= 1;
  register_A &= 0x0FFF;

  jumpCineRet_AA;

opLSLe_A0:
  register_A <<= 1; /* unit begin */
  register_PC++;
  if ( rom [ register_PC ] != 0xEC ) {
    goto opLSLe_A1; /* no more, do last one */
  }

  register_A <<= 1; /* unit begin */
  register_PC++;
  if ( rom [ register_PC ] != 0xEC ) {
    goto opLSLe_A1; /* no more, do last one */
  }

  register_A <<= 1; /* unit begin */
  register_PC++;
  if ( rom [ register_PC ] != 0xEC ) {
    goto opLSLe_A1; /* no more, do last one */
  }

  register_A <<= 1; /* unit begin */
  register_PC++;
  if ( rom [ register_PC ] != 0xEC ) {
    goto opLSLe_A1; /* no more, do last one */
  }

  register_A <<= 1; /* unit begin */
  register_PC++;
  if ( rom [ register_PC ] != 0xEC ) {
    goto opLSLe_A1; /* no more, do last one */
  }

  register_A <<= 1; /* unit begin */
  register_PC++;
  if ( rom [ register_PC ] != 0xEC ) {
    goto opLSLe_A1; /* no more, do last one */
  }

  register_A <<= 1;
  register_PC++;

opLSLe_A1:
  register_A &= 0x0FFF;

opLSLe_A2:
  cmp_new = temp_word;
  SETA0 ( register_A );
  cmp_old = register_A;

  temp_word += register_A;
  SETFC ( temp_word );

  register_A <<= 1;
  register_A &= 0x0FFF;

  jumpCineRet_AA;

#if 0 /* non-multiple-consecutive-LSLe's version */
  temp_word = 0x0CEC;                          /* data register */

  cmp_new = temp_word;                         /* magic value */
  SETA0 ( register_A );                        /* back up bit0 */
  cmp_old = register_A;                        /* store old acc */

  temp_word += register_A;                     /* add to acc */
  SETFC ( temp_word );                         /* store carry flag */
  register_A <<= 1;                            /* add regA to itself */
  register_A &= 0xFFF;                         /* toss excess bits */

  register_PC ++;                              /* bump PC */
  jumpCineRet_AA;
#endif

opLSLe_B_AA:
  temp_word = 0x0CEC;                          /* data register */

  cmp_new = temp_word;                         /* magic value */
  SETA0 ( register_A );                        /* back up bit0 */
  cmp_old = register_B;                        /* store old acc */

  temp_word += register_B;                     /* add to acc */
  SETFC ( temp_word );                         /* store carry flag */
  register_B <<= 1;                            /* add regA to itself */
  register_B &= 0xFFF;                         /* toss excess bits */

  register_PC ++;                              /* bump PC */
  jumpCineRet_AA;

opLSLf_A_AA:
opLSLf_AA_AA:
opLSLf_BB_AA:
  UNFINISHED ( "opLSLf 1\n" );
  jumpCineRet_AA;

opLSLf_B_AA:
  UNFINISHED ( "opLSLf 2\n" );
  jumpCineRet_AA;

opASRe_A_AA:
opASRe_AA_AA:
opASRe_BB_AA:
  /* agh! I dislike these silly 12bit processors :P */

  temp_word = 0xDED;

  register_PC ++;

  if ( rom [ register_PC ] == ( temp_word & 0xFF ) ) {
    goto opASRe_A0;
  }

  cmp_new = temp_word;

  SETA0 ( register_A );           /* store bit0 */
  SETFC ( register_A );

  cmp_old = register_A;

  register_A <<= 4; /* get sign bit */
  register_A = SAR(register_A,5);
  register_A &= 0xFFF;

  jumpCineRet_AA;

opASRe_A0:
  /* multiple ASRe's ... handle 'em in a batch, for efficiency
   */

  register_A <<= 4;
  register_A = SAR(register_A,1);

  register_PC++;
  if ( ! ( rom [ register_PC ] == ( temp_word & 0xFF ) ) ) {
    goto opASRe_A1;
  }                               /* end of unit */

  register_A = SAR(register_A,1);
  register_PC++;
  if ( ! ( rom [ register_PC ] == ( temp_word & 0xFF ) ) ) {
    goto opASRe_A1;
  }                               /* end of unit */

  register_A = SAR(register_A,1);
  register_PC++;
  if ( ! ( rom [ register_PC ] == ( temp_word & 0xFF ) ) ) {
    goto opASRe_A1;
  }                               /* end of unit */

  register_A = SAR(register_A,1);
  register_PC++;
  if ( ! ( rom [ register_PC ] == ( temp_word & 0xFF ) ) ) {
    goto opASRe_A1;
  }                               /* end of unit */

  register_A = SAR(register_A,1);
  register_PC++;
  if ( ! ( rom [ register_PC ] == ( temp_word & 0xFF ) ) ) {
    goto opASRe_A1;
  }                               /* end of unit */

  register_A = SAR(register_A,1);
  register_PC++;
  if ( ! ( rom [ register_PC ] == ( temp_word & 0xFF ) ) ) {
    goto opASRe_A1;
  }                               /* end of unit */

  register_A = SAR(register_A,1);
  register_PC++;
  
opASRe_A1:
  /* no more multiples left; finish off.
   */
  register_A >>= 4;

opASRe_A2:
  /* shift once with flags */
  cmp_new = temp_word;

  SETA0 ( register_A );           /* store bit0 */
  SETFC ( register_A );

  cmp_old = register_A;

  register_A <<= 4; /* get sign bit */
  register_A = SAR(register_A,5);
  register_A &= 0xFFF;

  jumpCineRet_AA;

opASRe_B_AA:
  register_PC++;

  if ( ( rom [ register_PC ] == 0xED ) &&
       ( rom [ register_PC + 1 ] == 0x57 ) )
  {
    goto opASRe_B0; /* another one follows, do multiples */
  }

  cmp_new = 0x0DED;
  SETA0 ( register_A );
  SETFC ( register_A );
  cmp_old = register_B;

  register_B <<= 4;
  register_B = SAR(register_B,5);
  register_B &= 0x0FFF;

  jumpCineRet_AA;

opASRe_B0:
  register_B <<= 4;  /* get sign bit */

  register_B = SAR(register_B,1);
  register_PC += 2;
  if ( ! ( ( rom [ register_PC ] == 0xED ) &&
	   ( rom [ register_PC + 1 ] == 0x57 ) ) )
  {
    goto opASRe_B1;
  }

  register_B = SAR(register_B,1);
  register_PC += 2;
  if ( ! ( ( rom [ register_PC ] == 0xED ) &&
	   ( rom [ register_PC + 1 ] == 0x57 ) ) )
  {
    goto opASRe_B1;
  }

  register_B = SAR(register_B,1);
  register_PC += 2;
  if ( ! ( ( rom [ register_PC ] == 0xED ) &&
	   ( rom [ register_PC + 1 ] == 0x57 ) ) )
  {
    goto opASRe_B1;
  }

  register_B = SAR(register_B,1);
  register_PC += 2;
  if ( ! ( ( rom [ register_PC ] == 0xED ) &&
	   ( rom [ register_PC + 1 ] == 0x57 ) ) )
  {
    goto opASRe_B1;
  }

  register_B = SAR(register_B,1);
  register_PC += 2;
  if ( ! ( ( rom [ register_PC ] == 0xED ) &&
	   ( rom [ register_PC + 1 ] == 0x57 ) ) )
  {
    goto opASRe_B1;
  }

  register_B = SAR(register_B,1);
  register_PC += 2;
  if ( ! ( ( rom [ register_PC ] == 0xED ) &&
	   ( rom [ register_PC + 1 ] == 0x57 ) ) )
  {
    goto opASRe_B1;
  }

  register_B = SAR(register_B,1);
  register_PC += 2;

opASRe_B1:
  register_B >>= 4; /* fix register */

opASRe_B2:
  cmp_new = 0x0DED;
  SETA0 ( register_A );
  SETFC ( register_A );
  cmp_old = register_B;

  register_B <<= 4;
  register_B = SAR(register_B,5);
  register_B &= 0x0FFF;

  jumpCineRet_AA;

opASRf_A_AA:
opASRf_AA_AA:
opASRf_BB_AA:
  UNFINISHED ( "opASRf 1\n" );
  jumpCineRet_AA;

opASRf_B_AA:
  UNFINISHED ( "opASRf 2\n" );
  jumpCineRet_AA;

opASRDe_A_AA:
opASRDe_AA_AA:
opASRDe_BB_AA:
  /* Arithmetic shift right of D (A+B) .. B is high (sign bits).
   * divide by 2, but leave the sign bit the same. (ie: 1010 -> 1001)
   */
  temp_word = 0x0EEE;
  register_PC++;

  if ( rom [ register_PC ] == ( temp_word & 0xFF ) ) {
    goto opASRDe_A0; /* multiples, do the batch */
  }

  cmp_new = temp_word;          /* save new acc value */
  SETA0 ( register_A & 0xFF );  /* save old accA bit0 */
  cmp_old = register_A;         /* save old acc */

  temp_word += register_A;
  SETFC ( temp_word );

  register_A <<= 4;
  register_B <<= 4;

  temp_word_2 = ( register_B >> 4 ) << 15;
  register_B = SAR(register_B,5);
  register_A = ( register_A >> 1 ) | temp_word_2;
  register_A >>= 4;

  register_B &= 0x0FFF;
  jumpCineRet_AA;

opASRDe_A0:
  register_A <<= 4;
  register_B <<= 4;

  temp_word_2 = ( register_B >> 4 ) << 15;
  register_B = SAR(register_B,5);
  register_A = ( register_A >> 1 ) | temp_word_2;

  register_PC++;
  if ( rom [ register_PC ] != ( temp_word & 0xFF ) ) {
    goto opASRDe_A1; /* no more, do last one */
  }

  temp_word_2 = ( register_B & 0x01 ) << 15;
  register_B = SAR(register_B,1);
  register_A = ( register_A >> 1 ) | temp_word_2;
  register_PC++;
  if ( rom [ register_PC ] != ( temp_word & 0xFF ) ) {
    goto opASRDe_A1; /* no more */
  }

  temp_word_2 = ( register_B & 0x01 ) << 15;
  register_B = SAR(register_B,1);
  register_A = ( register_A >> 1 ) | temp_word_2;
  register_PC++;
  if ( rom [ register_PC ] != ( temp_word & 0xFF ) ) {
    goto opASRDe_A1; /* no more */
  }

  temp_word_2 = ( register_B & 0x01 ) << 15;
  register_B = SAR(register_B,1);
  register_A = ( register_A >> 1 ) | temp_word_2;
  register_PC++;
  if ( rom [ register_PC ] != ( temp_word & 0xFF ) ) {
    goto opASRDe_A1; /* no more */
  }

  temp_word_2 = ( register_B & 0x01 ) << 15;
  register_B = SAR(register_B,1);
  register_A = ( register_A >> 1 ) | temp_word_2;
  register_PC++;
  if ( rom [ register_PC ] != ( temp_word & 0xFF ) ) {
    goto opASRDe_A1; /* no more */
  }

  temp_word_2 = ( register_B & 0x01 ) << 15;
  register_B = SAR(register_B,1);
  register_A = ( register_A >> 1 ) | temp_word_2;
  register_PC++;
  if ( rom [ register_PC ] != ( temp_word & 0xFF ) ) {
    goto opASRDe_A1; /* no more */
  }

  temp_word_2 = ( register_B & 0x01 ) << 15;
  register_B = SAR(register_B,1);
  register_A = ( register_A >> 1 ) | temp_word_2;

  register_PC++;

opASRDe_A1:  /* do last shift with flags */
  register_A >>= 4;

  cmp_new = temp_word;
  SETA0 ( register_A & 0xFF );
  cmp_old = register_A;

  temp_word += register_A;
  SETFC ( temp_word );

  register_A <<= 4;
  temp_word_2 = ( register_B & 0x01 ) << 15;
  register_B = SAR(register_B,1);
  register_A = ( register_A >> 1 ) | temp_word_2;
  register_A >>= 4;

  register_B &= 0x0FFF;
  jumpCineRet_AA;

opASRDe_B_AA:
  register_PC++;
  temp_word = 0x0EEE;
  cmp_new = temp_word;
  SETA0 ( register_A & 0xFF );
  cmp_old = register_B;

  temp_word += register_B;
  SETFC ( temp_word );
  register_B <<= 4;
  register_B = SAR(register_B,5);
  register_B &= 0x0FFF;

  jumpCineRet_AA;

opASRDf_A_AA:
opASRDf_AA_AA:
opASRDf_BB_AA:
  UNFINISHED ( "opASRDf 1\n" );
  jumpCineRet_AA;

opASRDf_B_AA:
  UNFINISHED ( "opASRDf 2\n" );
  jumpCineRet_AA;

opLSLDe_A_AA:
opLSLDe_AA_AA:
opLSLDe_BB_AA:
  /* LSLDe -- Left shift through both accumulators; lossy in middle. */

  temp_word = 0x0FEF;

  register_PC++;
  if ( rom [ register_PC ] == ( temp_word & 0xFF ) ) {
    goto opLSLDe_A0; /* multiples.. go to it. */
  }

  cmp_new = temp_word;
  SETA0 ( register_A );
  cmp_old = register_A;

  temp_word += register_A;
  SETFC ( temp_word );
  register_A <<= 1;                             /* logical shift left */
  register_A &= 0xFFF;

  register_B <<= 1;
  register_B &= 0xFFF;

  jumpCineRet_AA;

opLSLDe_A0:
  register_A <<= 1;
  register_B <<= 1;

  register_PC ++;
  if ( ! ( rom [ register_PC ] == ( temp_word & 0xFF ) ) ) {
    goto opLSLDe_A1; /* nope, go do last one */
  }

  register_A <<= 1;
  register_B <<= 1;

  register_PC ++;
  if ( ! ( rom [ register_PC ] == ( temp_word & 0xFF ) ) ) {
    goto opLSLDe_A1; /* nope, go do last one */
  }

  register_A <<= 1;
  register_B <<= 1;

  register_PC ++;
  if ( ! ( rom [ register_PC ] == ( temp_word & 0xFF ) ) ) {
    goto opLSLDe_A1; /* nope, go do last one */
  }

  register_A <<= 1;
  register_B <<= 1;

  register_PC ++;
  if ( ! ( rom [ register_PC ] == ( temp_word & 0xFF ) ) ) {
    goto opLSLDe_A1; /* nope, go do last one */
  }

  register_A <<= 1;
  register_B <<= 1;

  register_PC ++;
  if ( ! ( rom [ register_PC ] == ( temp_word & 0xFF ) ) ) {
    goto opLSLDe_A1; /* nope, go do last one */
  }

  register_A <<= 1;
  register_B <<= 1;

  register_PC ++;
  if ( ! ( rom [ register_PC ] == ( temp_word & 0xFF ) ) ) {
    goto opLSLDe_A1; /* nope, go do last one */
  }

  register_A <<= 1;
  register_B <<= 1;

  register_PC++;

opLSLDe_A1:
  register_A &= 0xFFF;
  register_B &= 0xFFF;

  cmp_new = temp_word;
  SETA0 ( register_A );
  cmp_old = register_A;

  temp_word += register_A;
  SETFC ( temp_word );
  register_A <<= 1;                             /* logical shift left */
  register_A &= 0xFFF;

  register_B <<= 1;
  register_B &= 0xFFF;

  jumpCineRet_AA;

opLSLDe_B_AA:
  UNFINISHED ( "opLSLD 1\n" );
  jumpCineRet_AA;

opLSLDf_A_AA:
opLSLDf_AA_AA:
opLSLDf_BB_AA:
  /* LSLDf */

  temp_word = 0x0FFF;

  register_PC++;
  if ( rom [ register_PC ] == ( temp_word & 0xFF ) ) {
    goto opLSLDf_A0; /* do multiple batches */
  }

  cmp_new = temp_word;
  SETA0 ( register_A );
  cmp_old = register_A;

  temp_word += register_A;
  SETFC ( temp_word );

  register_A <<= 1;
  register_A &= 0x0FFF;

  register_B <<= 1;
  register_B &= 0x0FFF;

  jumpCineRet_AA;

opLSLDf_A0:

  register_A <<= 1;  /* unit begin */
  register_B <<= 1;
  register_PC++;
  if ( rom [ register_PC ] != ( temp_word & 0xFF ) ) {
    goto opLSLDf_A1;
  }                  /* unit end */

  register_A <<= 1;  /* unit begin */
  register_B <<= 1;
  register_PC++;
  if ( rom [ register_PC ] != ( temp_word & 0xFF ) ) {
    goto opLSLDf_A1;
  }                  /* unit end */

  register_A <<= 1;  /* unit begin */
  register_B <<= 1;
  register_PC++;
  if ( rom [ register_PC ] != ( temp_word & 0xFF ) ) {
    goto opLSLDf_A1;
  }                  /* unit end */

  register_A <<= 1;  /* unit begin */
  register_B <<= 1;
  register_PC++;
  if ( rom [ register_PC ] != ( temp_word & 0xFF ) ) {
    goto opLSLDf_A1;
  }                  /* unit end */

  register_A <<= 1;  /* unit begin */
  register_B <<= 1;
  register_PC++;
  if ( rom [ register_PC ] != ( temp_word & 0xFF ) ) {
    goto opLSLDf_A1;
  }                  /* unit end */

  register_A <<= 1;  /* unit begin */
  register_B <<= 1;
  register_PC++;
  if ( rom [ register_PC ] != ( temp_word & 0xFF ) ) {
    goto opLSLDf_A1;
  }                  /* unit end */

  register_A <<= 1;
  register_B <<= 1;

  register_PC++;

opLSLDf_A1:
  register_A &= 0x0FFF;
  register_B &= 0x0FFF;

  cmp_new = temp_word;
  SETA0 ( register_A );
  SETFC ( register_A );
  cmp_old = register_A;

  temp_word += register_A;
  SETFC ( temp_word );

  register_A <<= 1;
  register_A &= 0x0FFF;

  register_B <<= 1;
  register_B &= 0x0FFF;

  jumpCineRet_AA;

opLSLDf_B_AA: /* not 'the same' as the A->AA version above */

  register_PC++;

  temp_word = 0x0FFF;
  cmp_new = temp_word;
  SETA0 ( register_A );
  cmp_old = register_B;

  temp_word += register_B;
  SETFC ( temp_word );

  register_B <<= 1;
  register_B &= 0x0FFF;

  jumpCineRet_AA;

opJMP_A_A:
opJMP_AA_A:
opJMP_BB_A:
  /* simple jump; change PC and continue..
   */

  /* Use 0xF000 so as to keep the current page, since it may well
   * have been changed with JPP.
   */
  register_PC = ( register_PC & 0xF000 ) + register_J;  /* pick up new PC */
  jumpCineRet_A;

opJMP_B_BB:
  register_PC = ( register_PC & 0xF000 ) + register_J;  /* pick up new PC */
  jumpCineRet_BB;

opJEI_A_A:
opJEI_AA_A:
opJEI_BB_A:

  if ( ! ( ioOutputs & 0x80 ) ) {
    goto opja1;
  }

  if ( ( FromX - JoyX ) > 0 ) {
    goto opJMP_A_A;
  }

  register_PC ++;   /* increment PC */
  jumpCineRet_A;

opja1:

  if ( ( FromX - JoyY ) > 0 ) {
    goto opJMP_A_A;
  }

  register_PC++;
  jumpCineRet_A;

opJEI_B_BB:

  if ( ! ( ioOutputs & 0x80 ) ) {
    goto opjbb1;
  }

  if ( ( FromX - JoyX ) > 0 ) {
    goto opJMP_B_BB;
  }

  register_PC ++;   /* increment PC */
  jumpCineRet_BB;

opjbb1:

  if ( ( FromX - JoyY ) > 0 ) {
    goto opJMP_B_BB;
  }

  register_PC++;
  jumpCineRet_BB;

opJEI_A_B:
opJEI_AA_B:
opJEI_BB_B:

  if ( ! ( ioOutputs & 0x80 ) ) {
    goto opjb1;
  }

  if ( ( FromX - JoyX ) > 0 ) {
    goto opJMP_A_B;
  }

  register_PC ++;   /* increment PC */
  jumpCineRet_B;

opjb1:

  if ( ( FromX - JoyY ) > 0 ) {
    goto opJMP_A_B;
  }

  register_PC++;
  jumpCineRet_B;

opJMI_A_A:
  /* previous instruction was not an ACC instruction, nor was the
   * instruction twice back a USB, therefore minus flag test the
   * current A-reg
   */

  /* negative acc? */
  if ( register_A & 0x800 ) {
    goto opJMP_A_A;  /* yes -- do jump */
  }

  register_PC ++;   /* increment PC */
  jumpCineRet_A;

opJMI_AA_A:
  /* previous acc negative? Jump if so... */
  if ( cmp_old & 0x800 ) {
    goto opJMP_AA_A;
  }
  register_PC++;
  jumpCineRet_A;

opJMI_BB_A:
  if ( register_B & 0x800 ) {
    goto opJMP_BB_A;
  }
  register_PC++;
  jumpCineRet_A;

opJMI_B_BB:
  if ( register_A & 0x800 ) {
    goto opJMP_B_BB;
  }
  register_PC++;
  jumpCineRet_BB;

opJLT_A_A:
opJLT_AA_A:
opJLT_BB_A:
  /* jump if old acc equals new acc */

  if ( cmp_new < cmp_old ) {
    goto opJMP_A_A;
  }

  register_PC ++;
  jumpCineRet_A;

opJLT_B_BB:
  if ( cmp_new < cmp_old ) {
    goto opJMP_B_BB;
  }
  register_PC ++;
  jumpCineRet_BB;

opJEQ_A_A:
opJEQ_AA_A:
opJEQ_BB_A:
  /* jump if equal */

  if ( cmp_new == cmp_old ) {
    goto opJMP_A_A;
  }
  register_PC++;                            /* bump PC */
  jumpCineRet_A;

opJEQ_B_BB:

  if ( cmp_new == cmp_old ) {
    goto opJMP_B_BB;
  }
  register_PC++;                            /* bump PC */
  jumpCineRet_BB;

opJA0_A_A:
opJA0_AA_A:
opJA0_BB_A:

  if ( acc_a0 & 0x01 ) {
    goto opJMP_A_A;
  }

  register_PC ++;                               /* bump PC */
  jumpCineRet_A;

opJA0_B_BB:
  if ( acc_a0 & 0x01 ) {
    goto opJMP_B_BB;
  }

  register_PC ++;                               /* bump PC */
  jumpCineRet_BB;

opJNC_A_A:
opJNC_AA_A:
opJNC_BB_A:

  if ( ! ( GETFC() & 0xF0 ) ) {
    goto opJMP_A_A; /* no carry, so jump */
  }
  register_PC ++;

  jumpCineRet_A;

opJNC_B_BB:
  if ( ! ( GETFC() & 0xF0 ) ) {
    goto opJMP_B_BB; /* no carry, so jump */
  }
  register_PC ++;

  jumpCineRet_BB;

opJDR_A_A:
opJDR_AA_A:
opJDR_BB_A:
  /*
   * ; Calculate number of cycles executed since
   * ; last 'VDR' instruction, add two and use as
   * ; cycle count, never branch
   */
  register_PC ++;
  jumpCineRet_A;

opJDR_B_BB:
  /*
   * ; Calculate number of cycles executed since
   * ; last 'VDR' instruction, add two and use as
   * ; cycle count, never branch
   */
  register_PC ++;
  jumpCineRet_BB;

opNOP_A_A:
opNOP_AA_A:
opNOP_BB_A:
  register_PC++;
  jumpCineRet_A;

opNOP_B_BB:
  register_PC++;
  jumpCineRet_BB;

opJPP32_A_B:
opJPP32_AA_B:
opJPP32_BB_B:
  /* ; 00 = Offset 0000h
   * ; 01 = Offset 1000h
   * ; 02 = Offset 2000h
   * ; 03 = Offset 3000h
   * ; 04 = Offset 4000h
   * ; 05 = Offset 5000h
   * ; 06 = Offset 6000h
   * ; 07 = Offset 7000h
   */
  temp_word = ( register_P & 0x07 ) << 12;  /* rom offset */
  register_PC = register_J + temp_word;
  jumpCineRet_B;

opJPP32_B_BB:
  temp_word = ( register_P & 0x07 ) << 12;  /* rom offset */
  register_PC = register_J + temp_word;
  jumpCineRet_BB;

opJPP16_A_B:
opJPP16_AA_B:
opJPP16_BB_B:
  /* ; 00 = Offset 0000h
   * ; 01 = Offset 1000h
   * ; 02 = Offset 2000h
   * ; 03 = Offset 3000h
   */
  temp_word = ( register_P & 0x03 ) << 12;  /* rom offset */
  register_PC = register_J + temp_word;
  jumpCineRet_B;

opJPP16_B_BB:
  temp_word = ( register_P & 0x03 ) << 12;  /* rom offset */
  register_PC = register_J + temp_word;
  jumpCineRet_BB;

opJMP_A_B:
  register_PC = ( register_PC & 0xF000 ) + register_J;  /* pick up new PC */
  jumpCineRet_B;

opJPP8_A_B:
opJPP8_AA_B:
opJPP8_BB_B:
  /* "long jump"; combine P and J to jump to a new far location (that can
   * be more than 12 bits in address). After this jump, further jumps
   * are local to this new page.
   */
  temp_word = ( ( register_P & 0x03 ) - 1 ) << 12;  /* rom offset */
  register_PC = register_J + temp_word;
  jumpCineRet_B;

opJPP8_B_BB:
  temp_word = ( ( register_P & 0x03 ) - 1 ) << 12;  /* rom offset */
  register_PC = register_J + temp_word;
  jumpCineRet_BB;

opJMI_A_B:
  if ( register_A & 0x800 ) {
    goto opJMP_A_B;
  }
  register_PC++;
  jumpCineRet_B;

opJMI_AA_B:
  UNFINISHED ( "opJMI 3\n" );
  jumpCineRet_B;

opJMI_BB_B:
  UNFINISHED ( "opJMI 4\n" );
  jumpCineRet_B;

opJLT_A_B:
opJLT_AA_B:
opJLT_BB_B:
  if ( cmp_new < cmp_old ) {
    goto opJMP_A_B;
  }
  register_PC ++;
  jumpCineRet_B;

opJEQ_A_B:
opJEQ_AA_B:
opJEQ_BB_B:
  if ( cmp_new == cmp_old ) {
    goto opJMP_A_B;
  }
  register_PC++;                            /* bump PC */
  jumpCineRet_B;

opJA0_A_B:
opJA0_AA_B:
opJA0_BB_B:

  if ( GETA0() & 0x01 ) {
    goto opJMP_A_B;
  }
  register_PC++;

  jumpCineRet_B;

opJNC_A_B:
opJNC_AA_B:
opJNC_BB_B:

  if ( ! ( GETFC() & 0x0F0 ) ) {
    goto opJMP_A_B; /* if no carry, jump */
  }
  register_PC++;

  jumpCineRet_B;

opJDR_A_B:
opJDR_AA_B:
opJDR_BB_B:
  /* register_PC++; */
  fprintf ( stderr, "The hell? No PC incrementing?\n" );
  jumpCineRet_B;

  /* NOP */
opNOP_A_B:
opNOP_AA_B:
opNOP_BB_B:
  register_PC++;                          /* NOP; bump PC only */
  jumpCineRet_B;

opLLT_A_AA:
opLLT_AA_AA:
opLLT_BB_AA:

  register_PC++;
  temp_byte = 0;

opLLTa1:
  temp_word = register_A >> 8;         /* register_A's high bits */
  temp_word &= 0x0A;                   /* only want PA11 and PA9 */

  if ( ! temp_word ) {
    goto opLLTa2;                      /* zero, no mismatch */
  }

  temp_word ^= 0x0A;                   /* flip the bits */

  if ( temp_word ) {
    goto opLLTa4;                      /* if not zero, mismatch found */
  }

opLLTa2:
  temp_word = register_B >> 8;         /* regB's top bits */
  temp_word &= 0x0A;                   /* only want SA11 and SA9 */

  if ( ! temp_word ) {
    goto opLLTa3;                      /* if zero, no mismatch */
  }

  temp_word ^= 0x0A;                   /* flip bits */

  if ( temp_word ) {
    goto opLLTa4;                      /* if not zero, mismatch found */
  }

opLLTa3:
  register_A <<= 1;                    /* shift regA */
  register_B <<= 1;                    /* shift regB */

  temp_byte ++;
  if ( temp_byte ) {
    goto opLLTa1;                      /* try again */
  }
  jumpCineRet_AA;

opLLTa4:
  vgShiftLength = temp_byte;
  register_A &= 0x0FFF;
  register_B &= 0x0FFF;

opLLTaErr:
  jumpCineRet_AA;

opLLT_B_AA:
  UNFINISHED ( "opLLT 1\n" );
  jumpCineRet_AA;

opVIN_A_A:
opVIN_AA_A:
opVIN_BB_A:
  /* set the starting address of a vector */

  FromX = register_A & 0xFFF;            /* regA goes to x-coord */
  FromY = register_B & 0xFFF;            /* regB goes to y-coord */

  register_PC ++;                             /* bump PC */
  jumpCineRet_A;

opVIN_B_BB:

  FromX = register_A & 0xFFF;            /* regA goes to x-coord */
  FromY = register_B & 0xFFF;            /* regB goes to y-coord */

  register_PC ++;                             /* bump PC */
  jumpCineRet_BB;

opWAI_A_A:
opWAI_AA_A:
opWAI_BB_A:
  /* wait for a tick on the watchdog
   */
  bNewFrame = 1;
  register_PC ++;
  goto cineExit;

opWAI_B_BB:
  bNewFrame = 1;
  register_PC ++;
  goto cineExit;

opVDR_A_A:
opVDR_AA_A:
opVDR_BB_A:
  /* set ending points and draw the vector, or buffer for a later draw.
   */
  {
    int ToX = register_A & 0xFFF;
    int ToY = register_B & 0xFFF;

    /* shl 20, sar 20; this means that if the CCPU reg should be -ve,
     * we should be negative as well.. sign extended.
     */
    if ( FromX & 0x800 ) {
      FromX |= 0xFFFFF000;
    }
    if ( ToX & 0x800 ) {
      ToX |= 0xFFFFF000;
    }
    if ( FromY & 0x800 ) {
      FromY |= 0xFFFFF000;
    }
    if ( ToY & 0x800 ) {
      ToY |= 0xFFFFF000;
    }

    /* figure out the vector
     */

    ToX -= FromX;
    ToX = SAR(ToX,vgShiftLength);
    ToX += FromX;

    ToY -= FromY;
    ToY = SAR(ToY,vgShiftLength);
    ToY += FromY;

    /* do orientation flipping, etc.
     */
    if ( ! bFlipX ) {
      goto noFlipX1;
    }

    ToX = sdwGameXSize - ToX;
    FromX = sdwGameXSize - FromX;

  noFlipX1:
    if ( ! bFlipY ) {
      goto noFlipY1;
    }
    
    ToY = sdwGameYSize - ToY;
    FromY = sdwGameYSize - FromY;

  noFlipY1:
    FromX += sdwXOffset;
    ToX += sdwXOffset;

    FromY += sdwYOffset;
    ToY += sdwYOffset;

    /* check real coords */
    if ( ! bSwapXY ) {
      goto noSwapXY2;
    }

    temp_word = ToY;
    ToY = ToX;
    ToX = temp_word;

    temp_word = FromY;
    FromY = FromX;
    FromX = temp_word;

  noSwapXY2:
    /* render the line
     */

    CinemaVectorData ( FromX, FromY,
		       ToX, ToY, 
		       vgColour + 0x0F );

#if 0
    fprintf ( stderr, "LINE: %d,%d -> %d,%d (%d)\n",
	      FromX, FromY, ToX, ToY, vgColour + 0x0F );
#endif

  }

  register_PC++;
  jumpCineRet_A;

opVDR_B_BB:
  UNFINISHED ( "opVDR B 1\n" );
  jumpCineRet_BB;

/* some code needs to be changed based on the machine or switches set.
 * Instead of getting disorganized, I'll put the extra dispatchers
 * here. The main dispatch loop jumps here, checks options, and
 * redispatches to the actual opcode handlers.
 */

/* JPP series of opcodes
 */
tJPP_A_B:
  /* MSIZE -- 0 = 4k, 1 = 8k, 2 = 16k, 3 = 32k */
  switch ( ccpu_msize ) {
  case 0:
  case 1:
    goto opJPP8_A_B;
    break;
  case 2:
    goto opJPP16_A_B;
    break;
  case 3:
    goto opJPP32_A_B;
    break;
  }
  fprintf ( stderr, "Out of range JPP!\n" );

tJPP_AA_B:
  /* MSIZE -- 0 = 4k, 1 = 8k, 2 = 16k, 3 = 32k */
  switch ( ccpu_msize ) {
  case 0:
  case 1:
    goto opJPP8_AA_B;
    break;
  case 2:
    goto opJPP16_AA_B;
    break;
  case 3:
    goto opJPP32_AA_B;
    break;
  }
  fprintf ( stderr, "Out of range JPP!\n" );

tJPP_B_BB:
  /* MSIZE -- 0 = 4k, 1 = 8k, 2 = 16k, 3 = 32k */
  switch ( ccpu_msize ) {
  case 0:
  case 1:
    goto opJPP8_B_BB;
    break;
  case 2:
    goto opJPP16_B_BB;
    break;
  case 3:
    goto opJPP32_B_BB;
    break;
  }
  fprintf ( stderr, "Out of range JPP!\n" );

tJPP_BB_B:
  /* MSIZE -- 0 = 4k, 1 = 8k, 2 = 16k, 3 = 32k */
  switch ( ccpu_msize ) {
  case 0:
  case 1:
    goto opJPP8_BB_B;
    break;
  case 2:
    goto opJPP16_BB_B;
    break;
  case 3:
    goto opJPP32_BB_B;
    break;
  }
  fprintf ( stderr, "Out of range JPP!\n" );

/* JMI series of opcodes
 */

tJMI_A_B:
  if ( ccpu_jmi_dip ) {
    goto opJMI_A_B;
  } else {
    goto opJEI_A_B;
  }

tJMI_A_A:
  if ( ccpu_jmi_dip ) {
    goto opJMI_A_A;
  } else {
    goto opJEI_AA_B;
  }

tJMI_AA_B:
  if ( ccpu_jmi_dip ) {
    goto opJMI_AA_B;
  } else {
    goto opJEI_AA_B;
  }

tJMI_AA_A:
  if ( ccpu_jmi_dip ) {
    goto opJMI_AA_A;
  } else {
    goto opJEI_AA_A;
  }

tJMI_B_BB1:
  if ( ccpu_jmi_dip ) {
    goto opJMI_B_BB;
  } else {
    goto opJEI_B_BB;
  }

tJMI_B_BB2:
  if ( ccpu_jmi_dip ) {
    goto opJMI_B_BB;
  } else {
    goto opJEI_B_BB;
  }

tJMI_BB_B:
  if ( ccpu_jmi_dip ) {
    goto opJMI_BB_B;
  } else {
    goto opJEI_BB_B;
  }

tJMI_BB_A:
  if ( ccpu_jmi_dip ) {
    goto opJMI_BB_A;
  } else {
    goto opJEI_BB_A;
  }

/* OUT series of opcodes:
 * ccpu_monitor can be one of:
 * 1 -- 16-level colour
 * 2 -- 64-level colour
 * 3 -- War of the Worlds colour
 * other -- bi-level
 */
tOUT_A_A:
  switch ( ccpu_monitor ) {
  case 1:
    goto opOUT16_A_A;
    break;
  case 2:
    goto opOUT64_A_A;
    break;
  case 3:
    goto opOUTWW_A_A;
    break;
  default:
    goto opOUTbi_A_A;
  }
  fprintf ( stderr, "Out of range OUT!\n" );
  goto opOUTbi_A_A;

tOUT_AA_A:
  switch ( ccpu_monitor ) {
  case 1:
    goto opOUT16_AA_A;
    break;
  case 2:
    goto opOUT64_AA_A;
    break;
  case 3:
    goto opOUTWW_AA_A;
    break;
  default:
    goto opOUTbi_AA_A;
  }
  fprintf ( stderr, "Out of range OUT!\n" );
  goto opOUTbi_A_A;

tOUT_B_BB:
  switch ( ccpu_monitor ) {
  case 1:
    goto opOUT16_B_BB;
    break;
  case 2:
    goto opOUT64_B_BB;
    break;
  case 3:
    goto opOUTWW_B_BB;
    break;
  default:
    goto opOUTbi_B_BB;
  }
  fprintf ( stderr, "Out of range OUT!\n" );
  goto opOUTbi_A_A;

tOUT_BB_A:
  switch ( ccpu_monitor ) {
  case 1:
    goto opOUT16_BB_A;
    break;
  case 2:
    goto opOUT64_BB_A;
    break;
  case 3:
    goto opOUTWW_BB_A;
    break;
  default:
    goto opOUTbi_BB_A;
  }
  fprintf ( stderr, "Out of range OUT!\n" );
  goto opOUTbi_A_A;

  /* opcode exit point (so we can catch and do debugging, etc.
   */
cineExecBottom:

#ifdef CCPUDEBUG
  /* handle debugger */
  if ( ccpudebug ) {
    disassemble ( disassembly, sizeof ( disassembly ), original_PC );
    disasmdump ( disassembly );
  }
#endif

#ifdef CCPUBREAK
  /* make adding opcodes easier..
   */
  if ( ccpubreak ) {
    if ( original_PC == register_PC ) {
      repeat_count++;

      if ( repeat_count == 1 ) {
	repeat_count = 0;    /* in case we don't want it to exit.. */
	fprintf ( stderr,       "Opcode repetition caused by "
		  "unknown opcode %d and state %d?\n",
		  rom [ original_PC ], original_state );
	/* HACK */ // exit ( 0 );
      }

    }

  }
#endif

  /* the opcode code has set a state and done mischief with flags and
   * the programcounter; now jump back to the top and run through another
   * opcode.
   */
#ifndef CCPUSSTEP
  if ( bBailOut == FALSE ) {
	 goto cineExecTop;
  }
#endif

cineExit:
  /* return control to main system
   */
  dwElapsedTicks = 100000; /* some magic number */

  return ( 0x80000000 );
}

/* Reset C-CPU registers, flags, etc to default starting values
 */
void cineReset(void) {

  /* ensure the context has been set at least once */
  ASSERT ( rom );

  /* zero registers */
  register_PC = 0;
  register_A = 0;
  register_B = 0;
  register_I = 0;
  register_J = 0;
  register_P = 0;
  FromX = 0;
  FromY = 0;
  register_T = 0;

  /* zero flags */
  flag_C = 0;

  /* reset state */
  state = state_A;

  /* reset ram*/
	memset(ram, 0, sizeof(ram));

  /* reset internal state */
  cmp_old = 0;
  cmp_new = 0;
  SETA0 ( 0 );

}

void cineSetJMI(UINT8 j) {
  ccpu_jmi_dip = j;

  if ( ccpu_jmi_dip ) {
    fprintf ( stderr, "CCPU JMI Set: Yes.\n" );
  } else {
    fprintf ( stderr, "CCPU JMI Set: No.\n" );
  }

}

void cineSetMSize(UINT8 m) {
  ccpu_msize = m;

  switch ( m ) {
  case 0:
    fprintf ( stderr, "CCPU Address Space: 4k\n" );
    break;
  case 1:
    fprintf ( stderr, "CCPU Address Space: 8k\n" );
    break;
  case 2:
    fprintf ( stderr, "CCPU Address Space: 16k\n" );
    break;
  case 3:
    fprintf ( stderr, "CCPU Address Space: 32k\n" );
    break;
  default:
    fprintf ( stderr, "CCPU Address Space: Error\n" );
  }

}

void cineSetMonitor(UINT8 m) {
  ccpu_monitor = m;

  switch ( m ) {
  case 1:
    fprintf ( stderr, "CCPU Monitor: 16-colour\n" );
    break;
  case 2:
    fprintf ( stderr, "CCPU Monitor: 64-colour\n" );
    break;
  case 3:
    fprintf ( stderr, "CCPU Monitor: War-of-the-Worlds-colour\n" );
    break;
  default:
    fprintf ( stderr, "CCPU Monitor: bi-level-display\n" );
  }

}

void cSetContext(void *p) {
	CONTEXTCCPU *c = p;
  cmp_old = c -> accVal;
  cmp_new = c -> cmpVal;
  rom = c -> cineBase;
  SETA0 ( c -> pa0 );
  flag_C = c -> cFlag;
  register_PC = c -> eRegPC;
  register_A = c -> eRegA;
  register_B = c -> eRegB;
  register_I = c -> eRegI;
  register_J = c -> eRegJ;
  register_P = c -> eRegP;
  state = (CINESTATE) c -> eCState;

#if 0
  fprintf ( stderr, "romBase = %p\n", rom );
#endif

}

void cGetContext(void *p) {

	CONTEXTCCPU *c = p;

  c -> accVal = cmp_old;
  c -> cmpVal = cmp_new;
  c -> cineBase = rom;
  c -> pa0 = GETA0();
  c -> cFlag = GETFC();
  c -> eRegPC = register_PC;
  c -> eRegA = register_A;
  c -> eRegB = register_B;
  c -> eRegI = register_I;
  c -> eRegJ = register_J;
  c -> eRegP = register_P;
  c -> eCState = state;
}

UINT32 cineGetElapsedTicks ( UINT32 dwClearIt ) {
  UINT32 dwTemp;

  dwTemp = dwElapsedTicks;
  if ( dwClearIt ) {
	 dwElapsedTicks = 0;
  }

  return ( dwTemp );
}

void cineReleaseTimeslice ( void ) {
  bBailOut = TRUE;
}

UINT32 cGetContextSize(void) {
  return ( sizeof ( CONTEXTCCPU ) );
}

/* extra functions
 */

/* disassemble() appends a disassembly of the specified address to
 * the end of the specified text buffer.
 */
void disassemble ( char *buffer, unsigned int maxlen, unsigned int address ) {
#ifdef CCPUDEBUG
  char ambuffer [ 40 ];   /* text buffer for addressing mode values */
  char opbuffer [ 40 ];   /* text buffer for opcodes etc. */
  char flbuffer [ 80 ];   /* text buffer for flag dump */
  CINEBYTE opcode;
  CINEBYTE opsize = 1;
  int iter = 0;

  /* which opcode in opcode table? (includes immediate data) */
  opcode = rom [ address ];

  /* build addressing mode (value to opcode) */

	memset(ambuffer, 0, sizeof(ambuffer));
  switch ( amodes [ opcode_table [ opcode ].ot_amode ].ad_amode ) {

  /* 4-bit immediate value */
  case AIM4:
  case IM4:
    StrPrintF ( ambuffer, "#$%X", opcode & 0x0F );
    break;

  /* 4-bit extended immediate value */
  case AIM4X:
    StrPrintF ( ambuffer, "#$%X", ( opcode & 0x0F ) << 8 );
    break;

  /* 8-bit immediate value */
  case AIM8:
    StrPrintF ( ambuffer, "#$%02X", rom [ address + 1 ] );
    opsize ++; /* required second byte for value */
    break;

  /* [m] -- indirect through memory */
  case AINDM:
    StrPrintF ( ambuffer, "AINDM/Error" );
    break;

  /* [i] -- indirect through 'I' */
  case AIRG:
  case IRG:
    StrPrintF ( ambuffer, "[i]" );
    break;

  /* extended 12-bit immediate value */
  case AIMX:
    StrPrintF ( ambuffer, "AIMX/Error" );
    break;

  /* no special params */
  case ACC:
  case AXLT:
  case IMP:
    ambuffer [ 0 ] = '\0';
    break;

  /* 12-bit immediate value */
  case IM12:
    StrPrintF ( ambuffer, "#$%03X",
	      ( rom [ address ] & 0x0F ) +
	      ( rom [ address + 1 ] & 0xF0 ) +
	      ( ( rom [ address + 1 ] & 0x0F ) << 8 ) );
    opsize ++;
    break;

  /* display address of direct addressing modes */
  case ADIR:
  case DIR:
    StrPrintF ( ambuffer, "$%X", opcode & 0x0F );
    break;

  /* jump through J register */
  case JUMP:
    StrPrintF ( ambuffer, "<jump>" );
    break;

  /* extended jump */
  case JUMPX:
    StrPrintF ( ambuffer, "JUMPX/Error" );
    break;

  } /* switch on addressing mode */

  /* build flags dump */
  StrPrintF ( flbuffer,
	    "A=%03X B=%03X I=%03XZ J=%03X P=%X " \
	    "A0=%02X %02X N%X O%X %c%c%c%c",
	    register_A, register_B,
	    register_I << 1,               /* use the <<1 for Zonn style */
	    register_J, register_P, GETA0(),
	    GETFC() /* ? 'C' : 'c' */,
	    cmp_new, cmp_old,
	    ( ( state == state_A ) || ( state == state_AA ) ) ? 'A' : ' ',
	    ( state == state_AA ) ? 'A' : ' ',
	    ( ( state == state_B ) || ( state == state_BB ) ) ? 'B' : ' ',
	    ( state == state_BB ) ? 'B' : ' ' );

  /* build the final output string. Format is as follows:
   * ADDRESS: OPCODE-ROM-DUMP\tOPCODE VALUE\tFLAGS
   */

  /* append complete opcode ROM dump */
	memset(opbuffer, 0, sizeof(opbuffer));
  for ( iter = 0; iter < opsize; iter++ ) {
    StrPrintF ( opbuffer + ( iter * 3 ), " %02X", rom [ address + iter ] );
  }
 
  /* create final output */
  StrPrintF ( buffer, "%04X:%-6s : %-4s %-6s : %s",
	    address, opbuffer,
	    opcodes [ opcode_table [ opcode ].ot_opcode ].od_name,
	    ambuffer,
	    flbuffer );

#endif

  return;
}
